use std::{io, path::Path, str::Utf8Error};

use serde::{Deserialize, Serialize};
use serde_json::Value;
use thiserror::Error;
use tokio::{fs, process::Command};
use tracing::{error, info, warn};
use url::Url;

use crate::string::ToUnquotedString;

#[derive(Debug, Error)]
pub enum ChannelError {
    #[error("URL is an invalid YouTube channel URL")]
    InvalidUrl,
    #[error("Channel already exists in database")]
    AlreadyExists,
    #[error("IO error: {0}")]
    IOError(#[from] io::Error),
    #[error("UTF-8 error: {0}")]
    Utf8Error(#[from] Utf8Error),
    #[error("Invalid output from ID extraction function")]
    InvalidOutput,
    #[error("Failed to extract channel ID from '{0}'")]
    ExtractID(String),
    #[error("Could not serialize info JSON: {0}")]
    SerializeInfoJSON(#[from] serde_json::Error),
    #[error("Failed to parse value from key '{0}'")]
    JsonKey(String),
}

#[derive(Serialize, Deserialize, Clone)]
pub struct Channel {
    pub id: i64,
    pub url: String,
    pub youtube_id: String,
    pub name: String,
    pub handle_url: String,
    pub avatar_url: String,
    pub banner_url: String,
    pub description: String,
    pub subscribers: i64,
}

impl Channel {
    async fn yt_dlp_task(url: &str) -> Result<(), ChannelError> {
        let mut child = Command::new("yt-dlp")
            .args([
                "--write-info-json",
                "--skip-download",
                "--playlist-items",
                "1",
                "-v",
                "-o",
                "media/channels/tmp/tmp.%(ext)s",
                url,
            ])
            .spawn()?;
        info!("yt-dlp task invoked");
        child.wait().await?;
        info!("yt-dlp task completed successfully");

        Ok(())
    }

    /// Transforms a channel handle or username into a canonical channel ID.
    ///
    /// While it invariably returns a canonical ID even if the URL already contains one,
    /// it's not recommended to call this function when that is the case in order to
    /// prevent multiple calls to yt-dlp.
    async fn extract_id(url: &str) -> Result<String, ChannelError> {
        let output = Command::new("yt-dlp")
            .args([
                "--skip-download",
                "--playlist-items",
                "1",
                "--print",
                "channel_id",
                url,
            ])
            .output()
            .await?;

        if !output.status.success() {
            error!(
                "yt-dlp could not download channel. Does the URL point to a valid/existing channel?"
            );
            return Err(ChannelError::InvalidOutput);
        }

        let stdout = str::from_utf8(&output.stdout)?;
        let mut lines = stdout.lines().filter(|line| line.starts_with("UC"));

        lines.next().map_or_else(
            || Err(ChannelError::ExtractID(url.into())),
            |id| Ok(id.into()),
        )
    }

    pub async fn from_url(url: &Url, id: i64) -> Result<Self, ChannelError> {
        let url_path = url.path();

        if !url_path.starts_with("/@")
            && !url_path.starts_with("/c/")
            && !url_path.starts_with("/channel/")
        {
            return Err(ChannelError::InvalidUrl);
        }

        let url = url.as_str();
        let youtube_id = if let Some(id) = url_path.strip_prefix("/channel/") {
            if id.starts_with("UC") {
                id.into()
            } else {
                warn!("Channel ID contains unexpected format: {id}, extracting");
                Self::extract_id(url).await?
            }
        } else {
            warn!("URL is not canonical, extracting ID");
            Self::extract_id(url).await?
        };

        let dir = format!("media/channels/{youtube_id}");
        let file_stem = format!("{dir}/{youtube_id}");

        if Path::new(&dir).exists() {
            warn!("Channel already exists, skipping");
            return Err(ChannelError::AlreadyExists);
        }

        fs::create_dir(dir).await?;

        Self::yt_dlp_task(url).await?;
        let info: Value =
            serde_json::from_str(&fs::read_to_string("media/channels/tmp/tmp.info.json").await?)?;

        let get_info_value = |key: &str| {
            info.get(key)
                .ok_or_else(|| ChannelError::JsonKey(key.to_string()))
                .unwrap()
        };

        let url = get_info_value("channel_url").to_unquoted_string();
        let name = get_info_value("channel").to_unquoted_string();
        let handle_url = get_info_value("uploader_url").to_unquoted_string();

        let mut avatar_url = String::new();
        let mut banner_url = String::new();

        if let Some(thumbnails) = info.get("thumbnails").and_then(Value::as_array) {
            for thumbnail in thumbnails {
                if let Some(id) = thumbnail.get("id").and_then(Value::as_str) {
                    match id {
                        "avatar_uncropped" => {
                            avatar_url = thumbnail
                                .get("url")
                                .and_then(Value::as_str)
                                .unwrap_or_default()
                                .to_string();
                        }
                        "banner_uncropped" => {
                            banner_url = thumbnail
                                .get("url")
                                .and_then(Value::as_str)
                                .unwrap_or_default()
                                .to_string();
                        }
                        _ => {}
                    }
                }
            }
        } else {
            warn!("Channel {youtube_id} has no thumbnails!");
        }

        let description = get_info_value("description").to_unquoted_string();
        let subscribers = get_info_value("channel_follower_count")
            .as_i64()
            .unwrap_or_default();

        fs::rename(
            "media/channels/tmp/tmp.info.json",
            format!("{file_stem}.info.json"),
        )
        .await?;

        Ok(Self {
            id,
            url,
            youtube_id,
            name,
            handle_url,
            avatar_url,
            banner_url,
            description,
            subscribers,
        })
    }
}
